@@ -14,6 +14,8 @@ package com.sun.jna;
 
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
+import java.nio.DoubleBuffer;
+import java.nio.FloatBuffer;
 import java.nio.IntBuffer;
 import java.nio.LongBuffer;
 import java.nio.ShortBuffer;
@@ -61,6 +63,12 @@ public class ArgumentsMarshalTest extends TestCase {
         String returnStringArrayElement(String[] args, int which);
         WString returnWideStringArrayElement(WString[] args, int which);
         Pointer returnPointerArrayElement(Pointer[] args, int which);
+
+        public static class TestPointerType extends PointerType {
+            public TestPointerType() { }
+            public TestPointerType(Pointer p) { super(p); }
+        }
+        TestPointerType returnPointerArrayElement(TestPointerType[] args, int which);
         CheckFieldAlignment returnPointerArrayElement(CheckFieldAlignment.ByReference[] args, int which);
         int returnRotatedArgumentCount(String[] args);
 
@@ -75,20 +83,60 @@ public class ArgumentsMarshalTest extends TestCase {
         int fillInt16Buffer(short[] buf, int len, short value);
         int fillInt32Buffer(int[] buf, int len, int value);
         int fillInt64Buffer(long[] buf, int len, long value);
+        int fillFloatBuffer(float[] buf, int len, float value);
+        int fillDoubleBuffer(double[] buf, int len, double value);
 
         // ByteBuffer alternative definitions
         int fillInt8Buffer(ByteBuffer buf, int len, byte value);
         int fillInt16Buffer(ByteBuffer buf, int len, short value);
         int fillInt32Buffer(ByteBuffer buf, int len, int value);
         int fillInt64Buffer(ByteBuffer buf, int len, long value);
+        int fillFloatBuffer(ByteBuffer buf, int len, float value);
+        int fillDoubleBuffer(ByteBuffer buf, int len, double value);
         
-        // {Short,Int,Long}Buffer alternative definitions        
+        // {Short|Int|Long|,Float|Double}Buffer alternative definitions
         int fillInt16Buffer(ShortBuffer buf, int len, short value);
         int fillInt32Buffer(IntBuffer buf, int len, int value);
         int fillInt64Buffer(LongBuffer buf, int len, long value);
+        int fillFloatBuffer(FloatBuffer buf, int len, float value);
+        int fillDoubleBuffer(DoubleBuffer buf, int len, double value);
 
         // Nonexistent functions 
         boolean returnBooleanArgument(Object arg);
+
+        // Structure
+        class MinTestStructure extends Structure {
+            public int field;
+        }
+        Pointer testStructurePointerArgument(MinTestStructure s);
+
+        class VariableSizedStructure extends Structure {
+            public int length;
+            public byte[] buffer;
+            public VariableSizedStructure(String arg) {
+                length = arg.length() + 1;
+                buffer = new byte[length];
+                System.arraycopy(arg.getBytes(), 0, buffer, 0, arg.length());
+            }
+        }
+        String returnStringFromVariableSizedStructure(VariableSizedStructure s);
+        class CbStruct extends Structure {
+            public static interface TestCallback extends Callback {
+                int callback(int arg1, int arg2);
+            }
+            public TestCallback cb;
+        }
+        void setCallbackInStruct(CbStruct cbstruct);
+
+        // Union (by value)
+        class TestUnion extends Union implements Structure.ByValue {
+            public String f1;
+            public int f2;
+        }
+        interface UnionCallback extends Callback {
+            TestUnion invoke(TestUnion arg);
+        }
+        TestUnion testUnionByValueCallbackArgument(UnionCallback cb, TestUnion arg);
     }
 
     TestLibrary lib;
@@ -228,9 +276,12 @@ public class ArgumentsMarshalTest extends TestCase {
             return new Integer(value);
         }
     }
-    public void testNativeMappedArgument() {
-        NativeMappedLibrary lib = (NativeMappedLibrary)
+    protected NativeMappedLibrary loadNativeMappedLibrary() {
+        return (NativeMappedLibrary)
             Native.loadLibrary("testlib", NativeMappedLibrary.class);
+    }
+    public void testNativeMappedArgument() {
+        NativeMappedLibrary lib = loadNativeMappedLibrary();
         final int MAGIC = 0x12345678;
         Custom arg = new Custom(MAGIC);
         assertEquals("Argument not mapped", MAGIC, lib.returnInt32Argument(arg));
@@ -280,6 +331,10 @@ public class ArgumentsMarshalTest extends TestCase {
                      + "parameter type, not argument type",
                      struct.getPointer(),
                      lib.testStructurePointerArgument(struct));
+
+        struct = null;
+        assertNull("Null argument should be returned",
+                   lib.testStructurePointerArgument(struct));
     }
 
     public void testStructureByValueArgument() {
@@ -310,7 +365,6 @@ public class ArgumentsMarshalTest extends TestCase {
     public void testWriteStructureArrayArgumentMemory() {
         final int LENGTH = 10;
         TestLibrary.CheckFieldAlignment block = new TestLibrary.CheckFieldAlignment();
-        block.useMemory(new Memory(block.size() * LENGTH));
         TestLibrary.CheckFieldAlignment[] array = 
             (TestLibrary.CheckFieldAlignment[])block.toArray(LENGTH);
         for (int i=0;i < array.length;i++) {
@@ -465,6 +519,67 @@ public class ArgumentsMarshalTest extends TestCase {
         }
     }
     
+    public void testWrappedByteArrayArgument() {
+        byte[] array = new byte[1024];
+        ByteBuffer buf = ByteBuffer.wrap(array, 512, 512).slice();
+        final byte MAGIC = (byte)0xAB;
+        lib.fillInt8Buffer(buf, 512, MAGIC);
+        for (int i=0;i < array.length;i++) {
+            assertEquals("Bad value at index " + i,
+                         i < 512 ? 0 : MAGIC, array[i]);
+        }
+    }
+    public void testWrappedShortArrayArgument() {
+        short[] array = new short[1024];
+        ShortBuffer buf = ShortBuffer.wrap(array, 512, 512).slice();
+        final short MAGIC = (short)0xABED;
+        lib.fillInt16Buffer(buf, 512, MAGIC);
+        for (int i=0;i < array.length;i++) {
+            assertEquals("Bad value at index " + i,
+                         i < 512 ? 0 : MAGIC, array[i]);
+        }
+    }
+    public void testWrappedIntArrayArgument() {
+        int[] array = new int[1024];
+        IntBuffer buf  = IntBuffer.wrap(array, 512, 512).slice();
+        final int MAGIC = 0xABEDCF23;
+        lib.fillInt32Buffer(buf, 512, MAGIC);
+        for (int i=0;i < array.length;i++) {
+            assertEquals("Bad value at index " + i,
+                         i < 512 ? 0 : MAGIC, array[i]);
+        }
+    }
+    public void testWrappedLongArrayArguent() {
+        long[] array = new long[1024];
+        LongBuffer buf  = LongBuffer.wrap(array, 512, 512).slice();
+        final long MAGIC = 0x1234567887654321L;
+        lib.fillInt64Buffer(buf, 512, MAGIC);
+        for (int i=0;i < array.length;i++) {
+            assertEquals("Bad value at index " + i,
+                         i < 512 ? 0 : MAGIC, array[i]);
+        }
+    }
+    public void testWrappedFloatArrayArguent() {
+        float[] array = new float[1024];
+        FloatBuffer buf  = FloatBuffer.wrap(array, 512, 512).slice();
+        final float MAGIC = -118.625f;
+        lib.fillFloatBuffer(buf, 512, MAGIC);
+        for (int i=0;i < array.length;i++) {
+            assertEquals("Bad value at index " + i,
+                         i < 512 ? 0 : MAGIC, array[i]);
+        }
+    }
+    public void testWrappedDoubleArrayArguent() {
+        double[] array = new double[1024];
+        DoubleBuffer buf  = DoubleBuffer.wrap(array, 512, 512).slice();
+        final double MAGIC = -118.625;
+        lib.fillDoubleBuffer(buf, 512, MAGIC);
+        for (int i=0;i < array.length;i++) {
+            assertEquals("Bad value at index " + i,
+                         i < 512 ? 0 : MAGIC, array[i]);
+        }
+    }
+    
     public void testInvalidArgument() {
         try {
             lib.returnBooleanArgument(this);
@@ -490,25 +605,44 @@ public class ArgumentsMarshalTest extends TestCase {
     
     public void testPointerArrayArgument() {
         Pointer[] args = { 
-      		new NativeString(getName()).getPointer(),
+            new NativeString(getName()).getPointer(),
             null,
             new NativeString(getName()+"2").getPointer(),
         };
+        Pointer[] originals = new Pointer[args.length];
+        System.arraycopy(args, 0, originals, 0, args.length);
+
         assertEquals("Wrong value returned", args[0], lib.returnPointerArrayElement(args, 0));
         assertNull("Wrong value returned", lib.returnPointerArrayElement(args, 1));
         assertEquals("Wrong value returned", args[2], lib.returnPointerArrayElement(args, 2));
         assertNull("Native array should be null terminated", lib.returnPointerArrayElement(args, 3));
+
+        assertSame("Argument pointers should remain unmodified [0]",
+                   originals[0], args[0]);
+        assertSame("Argument pointers should remain unmodified [2]",
+                   originals[2], args[2]);
     }
 
+    public void testNativeMappedArrayArgument() {
+        TestLibrary.TestPointerType[] args = {
+            new TestLibrary.TestPointerType(new NativeString(getName()).getPointer()),
+            null,
+            new TestLibrary.TestPointerType(new NativeString(getName()+"2").getPointer()),
+        };
+        assertEquals("Wrong value returned", args[0], lib.returnPointerArrayElement(args, 0));
+        assertNull("Wrong value returned", lib.returnPointerArrayElement(args, 1));
+        assertEquals("Wrong value returned", args[2], lib.returnPointerArrayElement(args, 2));
+    };
+
     public void testStructureByReferenceArrayArgument() {
         CheckFieldAlignment.ByReference[] args = { 
             new CheckFieldAlignment.ByReference(),
             null,
             new CheckFieldAlignment.ByReference(),
         };
-        assertEquals("Wrong value returned", args[0], lib.returnPointerArrayElement(args, 0));
-        assertNull("Wrong value returned", lib.returnPointerArrayElement(args, 1));
-        assertEquals("Wrong value returned", args[2], lib.returnPointerArrayElement(args, 2));
+        assertEquals("Wrong value returned (0)", args[0], lib.returnPointerArrayElement(args, 0));
+        assertNull("Wrong value returned (1)", lib.returnPointerArrayElement(args, 1));
+        assertEquals("Wrong value returned (2)", args[2], lib.returnPointerArrayElement(args, 2));
         assertNull("Native array should be null terminated", lib.returnPointerArrayElement(args, 3));
     }
 
@@ -520,6 +654,64 @@ public class ArgumentsMarshalTest extends TestCase {
                      Arrays.asList(args));
     }
     
+    public void testReadFunctionPointerAsCallback() {
+        TestLibrary.CbStruct s = new TestLibrary.CbStruct();
+        assertNull("Function pointer field should be null", s.cb);
+        lib.setCallbackInStruct(s);
+        assertNotNull("Callback field not set", s.cb);
+    }
+
+    public void testCallProxiedFunctionPointer() {
+        TestLibrary.CbStruct s = new TestLibrary.CbStruct();
+        lib.setCallbackInStruct(s);
+        assertEquals("Proxy to native function pointer failed: " + s.cb,
+                     3, s.cb.callback(1, 2));
+    }
+
+    public void testVariableSizedStructureArgument() {
+        String EXPECTED = getName();
+        TestLibrary.VariableSizedStructure s =
+            new TestLibrary.VariableSizedStructure(EXPECTED);
+        assertEquals("Wrong string returned from variable sized struct",
+                     EXPECTED, lib.returnStringFromVariableSizedStructure(s));
+    }
+
+    public void testDisableAutoSynch() {
+        TestLibrary.MinTestStructure s = new TestLibrary.MinTestStructure();
+        final int VALUE = 42;
+        s.field = VALUE;
+        s.setAutoWrite(false);
+        lib.testStructurePointerArgument(s);
+        assertEquals("Auto write should be disabled", 0, s.field);
+
+        final int EXPECTED = s.field;
+        s.getPointer().setInt(0, VALUE);
+        s.setAutoRead(false);
+        lib.testStructurePointerArgument(s);
+        assertEquals("Auto read should be disabled", EXPECTED, s.field);
+    }
+
+    public void testUnionByValueCallbackArgument() throws Exception{
+        TestLibrary.TestUnion arg = new TestLibrary.TestUnion();
+        arg.setType(String.class);
+        final String VALUE = getName();
+        arg.f1 = VALUE;
+        final boolean[] called = { false };
+        final String[] cbvalue = { null };
+        TestLibrary.TestUnion result = lib.testUnionByValueCallbackArgument(new TestLibrary.UnionCallback() {
+                public TestLibrary.TestUnion invoke(TestLibrary.TestUnion v) {
+                    called[0] = true;
+                    v.setType(String.class);
+                    v.read();
+                    cbvalue[0] = v.f1;
+                    return v;
+                }
+            }, arg);
+        assertTrue("Callback not called", called[0]);
+        assertEquals("Incorrect callback union argument", VALUE, cbvalue[0]);
+        assertEquals("Union value not propagated", VALUE, result.getTypedValue(String.class));
+    }
+
     public static void main(java.lang.String[] argList) {
         junit.textui.TestRunner.run(ArgumentsMarshalTest.class);
     }
